#include <stdio.h>
#include <time.h>
#include <unistd.h>
#include <errno.h>

#include "synch.h"
#include "my_signal.h"
#define SEM_TIMEOUT 20*1000*1000

int get_lock(pthread_mutex_t *my_mutex){
  sigset_t old_sig_mask;
  pthread_sigmask(SIG_SETMASK, &sig_mask, &old_sig_mask);
  pthread_mutex_lock(my_mutex);
  pthread_sigmask(SIG_SETMASK, &old_sig_mask, NULL);
  return 0;
}

int release_lock(pthread_mutex_t *my_mutex){
  sigset_t old_sig_mask;
  pthread_sigmask(SIG_SETMASK, &sig_mask, &old_sig_mask);
  pthread_mutex_unlock(my_mutex);
  pthread_sigmask(SIG_SETMASK, &old_sig_mask, NULL);
  return 0;
}

int init_lock(pthread_mutex_t *my_mutex){
  return pthread_mutex_init(my_mutex,NULL);
}

int destroy_lock(pthread_mutex_t *my_mutex){
  return pthread_mutex_destroy(my_mutex);
}


int sem_V(sem_t *my_sem){
  sigset_t old_sig_mask;
  pthread_sigmask(SIG_SETMASK, &sig_mask, &old_sig_mask);
  sem_post(my_sem);
  pthread_sigmask(SIG_SETMASK, &old_sig_mask, NULL);
  return 0;
}

int sem_P(sem_t *my_sem){
  sigset_t old_sig_mask;
  pthread_sigmask(SIG_SETMASK, &sig_mask, &old_sig_mask);
  sem_wait(my_sem);
  pthread_sigmask(SIG_SETMASK, &old_sig_mask, NULL);
  return 0;
}

int sem_tryP(sem_t *my_sem){
  sigset_t old_sig_mask;
  int r;

  pthread_sigmask(SIG_SETMASK, &sig_mask, &old_sig_mask);
  r = sem_trywait(my_sem);
  pthread_sigmask(SIG_SETMASK, &old_sig_mask, NULL);
  return r;
}

int sem_getValue(sem_t my_sem){
  sigset_t old_sig_mask;
  int r = -1;

  pthread_sigmask ( SIG_SETMASK, &sig_mask, &old_sig_mask );
  sem_getvalue(&my_sem, &r);
  pthread_sigmask ( SIG_SETMASK, &old_sig_mask, NULL);
  return r;
}

int init_sem(sem_t *my_sem,int value){
  return sem_init(my_sem, 0, value);
}

int destroy_sem(sem_t *my_sem){
  return sem_destroy(my_sem);
}
