#include <semaphore.h>
#include <stdio.h>
#include <stdlib.h>

#include "tampon.h"
#include "errno.h"
#include "trame.h"
#include "synch.h"

struct tampon{
  int nb_cases;
  struct trame **tab;
  int nb_elm_courant;

  sem_t prod;
  sem_t cons;

  pthread_mutex_t mutex;
};

struct tampon *creer_tampon(int nb_cases){
  struct tampon *tampon = malloc(sizeof(struct tampon));

  tampon->nb_cases = nb_cases;
  tampon->tab = malloc(nb_cases * sizeof(struct trame *));

  init_sem(&tampon->prod, nb_cases);
  init_sem(&tampon->cons, 0);
  
  init_lock(&tampon->mutex);

  return tampon;
}

struct trame *tampon_enlever(struct tampon *tampon){
  int i;
  int seq_min = -1;
  int idx = -1;

  sem_P(&tampon->cons);
  get_lock(&tampon->mutex);

  for(i = 0; i < tampon->nb_cases; i++){
    if(tampon->tab[i] != NULL){
      seq_min = tampon->tab[i]->info.data.numero_trame;
      idx = i;
    }
  }
  for( ; i < tampon->nb_cases; i++){
    if(tampon->tab[i] != NULL && seq_min > tampon->tab[i]->info.data.numero_trame){
      seq_min = tampon->tab[i]->info.data.numero_trame;
      idx = i;
    }
  }

  void *elm = tampon->tab[idx];
  tampon->tab[idx] = NULL;
  tampon->nb_elm_courant --;

  release_lock(&tampon->mutex);
  sem_V(&tampon->prod);

  return elm;
}

struct trame *tampon_enlever_seq(struct tampon *tampon, int seq){
  void *elm = NULL;
  int i;

  sem_P(&tampon->cons);
  get_lock(&tampon->mutex);

  for(i = 0; i < tampon->nb_cases; i++){
    if(tampon->tab[i] != NULL && tampon->tab[i]->info.data.numero_trame == seq){
      elm = tampon->tab[i];
      tampon->tab[i] = NULL;
      tampon->nb_elm_courant --;
    }
  }

  release_lock(&tampon->mutex);
  sem_V(&tampon->prod);

  return elm;
}

struct trame *tampon_rechercher(struct tampon *tampon, int seq){
  void *elm = NULL;
  int i;

  if(sem_getValue(tampon->cons) > 0){
    get_lock(&tampon->mutex);

    for(i = 0; i < tampon->nb_cases; i++){
      if(tampon->tab[i] != NULL && tampon->tab[i]->info.data.numero_trame == seq){
        elm = tampon->tab[i];
      }
    }

    release_lock(&tampon->mutex);
    //    sem_V(&tampon->prod);
  }

  return elm;
}

struct trame *tampon_suivant(struct tampon *tampon){
  struct trame *t = NULL;
  int seq_min = -1;
  int idx = -1;
  int i;

  if(sem_getValue(tampon->cons) > 0){
    get_lock(&tampon->mutex);

    for(i = 0; i < tampon->nb_cases; i++){
      if(tampon->tab[i] != NULL && tampon->tab[i]->flag_emis == 0){
        seq_min = tampon->tab[i]->info.data.numero_trame;
        tampon->tab[i]->flag_emis = 1;
        idx = i;
      }
    }

    for( ; i < tampon->nb_cases; i++){
      if(tampon->tab[i] != NULL
         && tampon->tab[i]->flag_emis == 0
         && tampon->tab[i]->info.data.numero_trame < seq_min){

        seq_min = tampon->tab[i]->info.data.numero_trame;
        tampon->tab[i]->flag_emis = 1;
        idx = i;
      }
    }

    if(idx != -1)
      t = tampon->tab[idx];

    release_lock(&tampon->mutex);
  }

  return t;
}



void tampon_ajouter(struct tampon *tampon, struct trame *t){
  int i;

  sem_P(&tampon->prod);
  get_lock(&tampon->mutex);

  for(i = 0; i < tampon->nb_cases; i++){
    if(tampon->tab[i] == NULL){
      tampon->tab[i] = t;
      tampon->nb_elm_courant++;
      break;
    }
  }

  release_lock(&tampon->mutex);
  sem_V(&tampon->cons);
}

int tampon_essai_ajouter(struct tampon *tampon,  struct trame *t){
  int r = 0;
  int i;

  if(sem_tryP(&tampon->prod) == 0){
    get_lock(&tampon->mutex);
    for(i = 0; i < tampon->nb_cases; i++){
      if(tampon->tab[i] == NULL){
        tampon->tab[i] = t;
        tampon->nb_elm_courant++;

        r = 1;
        break;
      }
    }

    release_lock(&tampon->mutex);
    sem_V(&tampon->cons);
  }
  return r;
}

int tampon_nb_elm(struct tampon *tampon){
  int nb_elm_courant = 0;

  get_lock(&tampon->mutex);
  nb_elm_courant = tampon->nb_elm_courant;
  release_lock(&tampon->mutex);

  return nb_elm_courant;
}
