#include <pthread.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <stdlib.h>

#include <sys/socket.h>

#include <time.h>
#include <string.h>
#include <errno.h>

#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>


#include "tcp_like.h"
#include "tcp_like_remettre.h"

#include "perror.h"
#include "trame.h"
#include "tampon.h"
#include "timeout.h"
#include "my_signal.h"

#include "appli.h"
#include "synch.h"

#define PORT_S 3421
#define SRV_IP "127.0.0.1"

static int sock;

static int s_seq = 0;
int last_frame;
int nb_frames;
pthread_t tr, ts;
int end_reached;
int all_frames_received;
int still_work;
int fin_received;
int fin_sent;


static struct sockaddr_in si_other_s,si_other_r;

struct tampon *tampon_emission = NULL;
struct tampon *tampon_reception = NULL;

static void *protocole_reception(void *);
static void *protocole_emission(void *);

static void init_threads(void);
static void init_tampons(void);

static int flag_client;
static int cnx_nb = 0;
static int nb_sock = 0;
sem_t barrier;

void tcp_like_launch(int client){
  flag_client = client;
  all_frames_received=0;
  last_frame=0;
  end_reached=0;
  nb_frames=0;
  still_work=0;
  fin_received=0;
  fin_sent=0;

  init_tampons();
  timeout_init_layer();
  init_signals();
  init_sem(&barrier,0);
  init_trame();
  init_threads();
}

void tcp_like_exit(void){
  pthread_join(tr, NULL);
  pthread_join(ts, NULL);
  close(sock);
}

void tcp_like_accept(void){
  struct sockaddr_in si;
 
  if ((sock = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP)) == -1)
    perror_and_exit("socket - chec  l'ouverture");

  memset((char *) &si, sizeof(si), 0);
  si.sin_family = AF_INET;
  si.sin_port = htons(PORT_S);
  si.sin_addr.s_addr = htonl(INADDR_ANY);

  if (bind(sock, (struct sockaddr *) &si, sizeof(si)) == -1)
    perror_and_exit("Echec sur bind");

  nb_sock ++;

  return;
}

void tcp_like_connect(void){
  struct trame *t = NULL;
    
  if ((sock = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP)) == -1)
    perror_and_exit("socket");

  memset((char *) &si_other_s, sizeof(si_other_s), 0);
  si_other_s.sin_family = AF_INET;
  si_other_s.sin_port = htons(PORT_S);
  if (inet_aton(SRV_IP, &si_other_s.sin_addr)==0) {
    perror_and_exit("inet_aton() failed\n");
  }

  nb_sock ++;  
  t = creer_demande_cnx();
  // envoi du syn
  fprintf(stderr, "Sending SYN\n");
  
  emettre_trame(t, sock, &si_other_s);
  detruire_trame(t);

  return;
}


void tcp_like_send(void *buf, int taille){
  struct trame *trame = creer_trame_donnees(s_seq++, buf, taille);
  tampon_ajouter(tampon_emission, trame);
}


void tcp_like_recv(void *buf, int taille){
  struct trame *trame = tampon_enlever(tampon_reception);
  memcpy(buf, trame->info.data.data, taille);
  detruire_trame(trame);
}

void tcp_like_disconnect() {
  envoyer_trame_retardee(sock, &si_other_s);
  /* il faut faire la fermeture de la connexion avec FIN, FIN ACK etc ... */
  struct trame *t;
  fprintf(stderr, "Planning to send a FIN \n");
  t=creer_fin(s_seq);
  fprintf(stderr, "Waiting for %d frames\n",s_seq);
  emettre_trame(t, sock, &si_other_s);
  detruire_trame(t);
}

/**********************************************/
/**********************************************/
/**********************************************/

static void init_threads(void){
  if(pthread_create(&tr, NULL, protocole_reception, NULL))
    perror_and_exit("Echec sur la cration du thread de reception");

  if(pthread_create(&ts, NULL, appli, &flag_client))
    perror_and_exit("Echec sur la cration du thread de l'application");

  // le thread principal s'occupe de l'mission
  protocole_emission(NULL);
}

static void init_tampons(void){
  tampon_reception = creer_tampon(1);
  tampon_emission  = creer_tampon(1);
}



void tcp_like_remettre_trame(int n_trame){
  printf("Re-sending frame %d\n",n_trame);
  struct trame *t = tampon_rechercher(tampon_emission, n_trame);

  emettre_trame(t, sock, &si_other_s);
}



static void *protocole_reception(void *arg){
  struct trame *trame = NULL;
  int octets_recus = 0;
  int cur_nb_octets_lus = 0;
  int sockaddr_in_len = sizeof(struct sockaddr_in);
  struct trame *t = NULL;
  fd_set read_fd_set;
  struct timeval tmp_time;
  int i,tmp;
  /*int flag=0;*/
  
  while((!end_reached)||(tampon_nb_elm(tampon_emission)!=0)||(tampon_nb_elm(tampon_reception)!=0)||(!all_frames_received)){
    while(!nb_sock);
    
    FD_ZERO (&read_fd_set);
    FD_SET (sock,&read_fd_set);

    tmp_time.tv_sec=0;
    tmp_time.tv_usec=1000*2;

    i= my_select (sock);
    if(i<0){
      perror_and_exit("Erreur select");
    }else{
      if(i!=0){
	octets_recus = 0;
	cur_nb_octets_lus = 0;
	
	trame = creer_trame_de_reception();
	if(recvfrom(sock, trame + cur_nb_octets_lus, sizeof(struct trame)- cur_nb_octets_lus, 0, (struct sockaddr *)&si_other_r, (unsigned *)&sockaddr_in_len) == -1)
	  perror_and_exit("proto_recv - echec sur recvfrom");
	printf("Received packet from %s:%d\n",
	       inet_ntoa(si_other_r.sin_addr), ntohs(si_other_r.sin_port));
	if(memcmp(&si_other_s,&si_other_r,sizeof(si_other_r))!=0) {
	  memcpy(&si_other_s,&si_other_r,sizeof(si_other_r));
	}

	switch(trame->type){
	case data:
	  {
	    /* struct trame *trame_ack = creer_ack(...); */
	    fprintf(stderr, "Sending ACK (%d)\n", nb_frames);
	    /*emettre_trame(...);*/
	    break;
	  }
	case ack:
	  {
	    fprintf(stderr, "Removing timeout (%d)\n", trame->info.ack.numero_ack);
	    /* t = tampon_enlever_seq(...); */
	    break;
	  }
	case requete_connexion:
	  {
	    fprintf(stderr, "Connection request received, sending SYN\n");
	    t = creer_succes_cnx(cnx_nb);
	    cnx_nb++;
	    // si tout va bien -> envoi de syn sinon de rst
	    emettre_trame(t, sock, &si_other_s);
	    detruire_trame(t);
	    break;
	  }
	case connexion_succes:
	  {
	    if(trame->type == connexion_rejet){
	      perror_and_exit("Connection refused");
	    } else {
	      // envoi d'un ack
	      t = creer_sack(trame->info.connexion_succes.cnx_nb);
	      emettre_trame(t, sock, &si_other_s);
	      detruire_trame(t);
	      sem_V(&barrier);
	    }
	    break;
	  }
	case connexion_rejet:
	  break;
	case sack:
	  {
	    if(trame->info.sack.cnx_nb >= cnx_nb)
	      perror_and_exit("sack - non valide");
	    fprintf(stderr, "Connection established\n");
	    sem_V(&barrier);
	    break;
	  }
	case fin:
	  {
	    fprintf(stderr, "FIN received, sending a FIN ACK %d\n",trame->info.fin.seq_number);
	    last_frame=trame->info.fin.seq_number;
	    t = creer_fin_ack();
	    emettre_trame(t, sock, &si_other_s);
	    detruire_trame(t);
	    fin_received=1;
	    if(nb_frames==last_frame){
	      all_frames_received=1;
	    }
	    if((fin_sent)&&(fin_received)){
	      end_reached=1;
	    }
	    break;
	  }
	case fin_ack:
	  {
	    fprintf(stderr, "FIN ACK received\n");
	    fin_sent=1;
	    if(nb_frames==last_frame){
	      all_frames_received=1;
	    }
	    if((fin_sent)&&(fin_received)){
	      end_reached=1;
	    }
	    break;
	  }
	default:
	  perror_and_exit("Type de message reu non valide");
	}
      }
    }
  }
  printf("thread reception : je m'arrete \n");
  still_work=1;
  return NULL;
}

static void *protocole_emission(void *arg){
  struct trame *t = NULL;
  sem_P(&barrier);
  while((!end_reached)||(tampon_nb_elm(tampon_emission)!=0)||(tampon_nb_elm(tampon_reception)!=0)||(!still_work)){
    t = tampon_suivant(tampon_emission);
    if(t){
      printf("j'emets une trame (%d) \n",t->info.fin.seq_number); 
      emettre_trame(t, sock, &si_other_s);
    }else{
      if((!fin_sent)&&(fin_received)&&(tampon_nb_elm(tampon_emission)==0)&&(tampon_nb_elm(tampon_reception)==0)){
	t=creer_fin(s_seq);
	fprintf(stderr, "Waiting for %d frames\n",s_seq);
	emettre_trame(t, sock, &si_other_s);
	detruire_trame(t);
      }
    }
  }
  printf("thread emission : je m'arrete \n");
  return NULL;
}
