#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <arpa/inet.h>

#include "trame.h"
#include "reallife.h"
#include "perror.h"
#include "synch.h"
#include "timeout.h"

pthread_mutex_t sock_lock;
struct trame *trame_retardee;

void
init_trame() {
  
  init_lock(&sock_lock);
  trame_retardee = NULL;
}

struct trame *creer_trame_donnees(int numero_trame, void *buf, int taille){
  struct trame *t = malloc(sizeof(struct trame));
  memset(t, '0', sizeof(struct trame));

  t->type = data;
  t->flag_emis = 0;

  t->info.data.numero_trame = numero_trame;

  if(buf != NULL){
    memcpy(t->info.data.data, buf, taille);
  }

  return t;
}

struct trame *creer_trame_de_reception(void){
  struct trame *t = malloc(sizeof(struct trame));
  memset(t, '0', sizeof(struct trame));

  return t;
}

struct trame *creer_ack(int numero_ack){
  struct trame *t = malloc(sizeof(struct trame));
  memset(t, '0', sizeof(struct trame));

  t->type = ack;
  t->info.ack.numero_ack = numero_ack;
  return t;
}

struct trame *creer_demande_cnx(void){
  struct trame *t = malloc(sizeof(struct trame));
  memset(t, '0', sizeof(struct trame));

  t->type = requete_connexion;

  return t;
}



struct trame *creer_succes_cnx(int cnx_nb){
  struct trame *t = malloc(sizeof(struct trame));
  memset(t, '0', sizeof(struct trame));

  t->type = connexion_succes;
  t->info.connexion_succes.cnx_nb = cnx_nb;

  return t;
}

struct trame *creer_rejet_cnx(void){
  struct trame *t = malloc(sizeof(struct trame));
  memset(t, '0', sizeof(struct trame));

  t->type = connexion_rejet;
  return t;
}

struct trame *creer_sack(int cnx_nb){
  struct trame *t = malloc(sizeof(struct trame));
  memset(t, '0', sizeof(struct trame));

  t->type = sack;
  t->info.sack.cnx_nb = cnx_nb;

  return t;
}


struct trame *creer_fin(int seq_number){
  struct trame *t = malloc(sizeof(struct trame));
  memset(t, '0', sizeof(struct trame));
  t->info.fin.seq_number=seq_number;
  t->type = fin;

  return t;
}

struct trame *creer_fin_ack(void){
  struct trame *t = malloc(sizeof(struct trame));
  memset(t, '0', sizeof(struct trame));

  t->type = fin_ack;

  return t;
}

static void
envoyer(struct trame *trame, int sock, struct sockaddr_in *si_other) {
 int sockaddr_in_len = sizeof(struct sockaddr_in);
 int octets_envoyes = -1;

  get_lock(&sock_lock);
  if (trame->type == data) {
    fprintf(stderr, "Setting up timeout for frame %d\n", trame->info.data.numero_trame);
    timeout_set_alarm(trame->info.data.numero_trame, 0); /****** type ????****/
  }
  
  if ((octets_envoyes = sendto(sock, trame, sizeof(struct trame), 0, (struct sockaddr *)si_other, sockaddr_in_len)) == -1) {
    perror_and_exit("sendto()");
  }
  release_lock(&sock_lock);
  fprintf(stderr, " dst -> %s | dport -> %d\n",inet_ntoa(si_other->sin_addr), ntohs(si_other->sin_port));
}

/* Envoie la trame "trame" sur la socket "sock" au destinataire dont
   l'adresse est contenue par "si_other" */
void emettre_trame(struct trame *trame, int sock, struct sockaddr_in *si_other) {
  int roll; 
  if (real_life_is_on)
    roll = hasard();
  else
    roll = TOUT_EST_OK;

  if(trame_retardee!=NULL){
    if(memcmp(trame_retardee,trame,sizeof(struct trame))==0){
      trame_retardee=NULL;
    }
  }

  switch(roll) {
  case TRAME_DETRUITE:
    {
      if (trame->type == data || trame->type == ack) {
	fprintf(stderr, "Frame lost !\n");
	if(trame->type == data){
	  fprintf(stderr, "Setting up timeout for frame %d\n", trame->info.data.numero_trame);
	  timeout_set_alarm(trame->info.data.numero_trame, 0); /****** type ????****/
	}
      }else{
	envoyer(trame, sock, si_other);
      }
      break;
    }
  case TRAME_RETARDEE:
    {
      if(trame->type == data) {
	fprintf(stderr, "Delayed Frame !\n");
	envoyer_trame_retardee(sock, si_other);
	trame_retardee = trame;
	fprintf(stderr, "Setting up timeout for frame %d\n", trame->info.data.numero_trame);
	timeout_set_alarm(trame->info.data.numero_trame, 0); /****** type ????****/
	return;
      }

      envoyer(trame, sock, si_other);
      break;
    }
  case TOUT_EST_OK:
    {
      envoyer(trame, sock, si_other);
      break;
    }
  default:
    {
      fprintf(stderr, "Unexpected real-life event !\n");
      exit(EXIT_FAILURE);
    }
  }
  envoyer_trame_retardee(sock, si_other);
}

void
envoyer_trame_retardee(int sock, struct sockaddr_in *si_other) {
  if (trame_retardee) {
    envoyer(trame_retardee, sock, si_other);
    trame_retardee = NULL;
  }    
}

void detruire_trame(struct trame *trame){
  free(trame);
}
